#include <stdlib.h>
#include <stdio.h>
#include "lzo/lzoconf.h"
#include "lzo/lzo1x.h"

LZO_EXTERN(int)
lzo1x_999_compress_internal ( const lzo_bytep in , lzo_uint  in_len,
                                    lzo_bytep out, lzo_uintp out_len,
                                    lzo_voidp wrkmem,
                              const lzo_bytep dict, lzo_uint dict_len,
                                    lzo_callback_p cb,
                                    int try_lazy,
                                    lzo_uint good_length,
                                    lzo_uint max_lazy,
                                    lzo_uint nice_length,
                                    lzo_uint max_chain,
                                    lzo_uint32 flags );

unsigned int max_compressed_size(unsigned int uncompressed_size)
{
  return uncompressed_size + uncompressed_size / 16 + 64 + 3;
}

int compress(char *in, unsigned int in_len, char *out, unsigned int out_bufsize, unsigned int *out_len)
{
    int r;
    int lazy;
    const int max_try_lazy = 10;
    const lzo_uint big = 65536L;    /* can result in very slow compression */
    const lzo_uint32 flags = 0x1;
    lzo_bytep wrkmem;
    lzo_uint best_len = out_bufsize;
    int best_lazy = -1;
    lzo_uint orig_len;

    if ( out_bufsize < max_compressed_size(in_len) )
        return -1;

/*
 * Step 1: initialize the LZO library
 */
    if (lzo_init() != LZO_E_OK)
        return -2;

/*
 * Step 2: allocate the work-memory
 */
    wrkmem = (lzo_bytep) malloc(LZO1X_999_MEM_COMPRESS);
    if (wrkmem == NULL)
        return -3;

/*
 * Step 3: compress from `in' to `out' with LZO1X-999
 */
    for (lazy = 0; lazy <= max_try_lazy; lazy++)
    {
        *out_len = out_bufsize;
        r = lzo1x_999_compress_internal(in,in_len,out,out_len,wrkmem,
                                        NULL, 0, 0,
                                        lazy, big, big, big, big, flags);
        if (r != LZO_E_OK)
        {
            free(wrkmem);
            return -4;
        }
//        printf("LZO1X-999: lazy =%2d: %8lu -> %8lu\n",
//                lazy, (unsigned long) in_len, (unsigned long) out_len);
        if (*out_len < best_len)
        {
            best_len = *out_len;
            best_lazy = lazy;
        }
    }

/*
 * Step 4: compress data again using the best compressor found
 */
    *out_len = out_bufsize;
    r = lzo1x_999_compress_internal(in,in_len,out,out_len,wrkmem,
                                    NULL, 0, 0,
                                    best_lazy, big, big, big, big, flags);
    if (r != LZO_E_OK || *out_len != best_len)
    {
        free(wrkmem);
        return -5;
    }

/*
 * Step 5: optimize compressed data (compressed data is in `out' buffer)
 */
    orig_len = in_len;
    r = lzo1x_optimize(out,*out_len,in,&orig_len,NULL);
    if (r != LZO_E_OK || orig_len != in_len)
    {
        free(wrkmem);
        return -6;
    }
    free(wrkmem);
    
    return 0;
}

int __lzo_cdecl_main main(int argc, char *argv[])
{
    lzo_bytep in;
    lzo_uint in_len;

    lzo_bytep out;
    lzo_uint out_bufsize;
    lzo_uint out_len = 0;

    FILE *f;
    const char *progname = NULL;
    const char *in_name = NULL;
    const char *out_name = NULL;
    long l;
    int p;

    progname = argv[0];
    if (argc < 2 || argc > 3)
    {
        printf("\nLZO real-time data compression library (v%s, %s).\n",
               lzo_version_string(), lzo_version_date());
        printf("Copyright (C) 1996-2005 Markus Franz Xaver Johannes Oberhumer\nAll Rights Reserved.\n");
        printf("MiniLZO LZO1X-999-only compressor by Maxim, based on the PRECOMP2 example.\n");
        printf("usage: %s file [output-file]\n", progname);
        printf("wildcard usage: for %%f in (*.foo) do @\"%s\" \"%%f\" [output-file]\n", progname);
        exit(1);
    }
    in_name = argv[1];
    if (argc > 2) out_name = argv[2];

/*
 * Step 1: open the input file
 */
    f = fopen(in_name,"rb");
    if (f == NULL)
    {
        printf("%s: cannot open file %s\n", progname, in_name);
        exit(1);
    }
    fseek(f,0,SEEK_END);
    l = ftell(f);
    fseek(f,0,SEEK_SET);
    if (l <= 0)
    {
        printf("%s: %s: empty file\n", progname, in_name);
        fclose(f);
        exit(1);
    }
    in_len = (lzo_uint) l;
    out_bufsize = max_compressed_size(in_len);

/*
 * Step 2: allocate compression buffers and read the file
 */
    in = (lzo_bytep) malloc(in_len);
    out = (lzo_bytep) malloc(out_bufsize);
    if (in == NULL || out == NULL)
    {
        printf("%s: out of memory\n", progname);
        exit(1);
    }
    in_len = (lzo_uint) fread(in,1,in_len,f);
//    printf("%s: loaded file %s: %ld bytes\n", progname, in_name, (long) in_len);
    fclose(f);

/*
 * Step 3: compress from one buffer to the other
 */
    printf("Compressing %-24s (%8d bytes) to %-24s", in_name, in_len, (out_name?out_name:"null"));
    p = compress(in, in_len, out, out_bufsize, &out_len);
    if ( p == 0 )
    {
      printf(" (%8d bytes): %2d%% compression.\n", out_len, (int)((1-(float)out_len/in_len)*100));
    }
    else
    {
      printf("Error %d\n", p );
    }

/*
 * Step 4: write compressed data to a file
 */

    if (out_name && out_name[0])
    {
        f = fopen(out_name,"wb");
        if (f == NULL)
        {
            printf("%s: cannot open output file %s\n", progname, out_name);
            exit(1);
        }
        if (fwrite(out,1,out_len,f) != out_len || fclose(f) != 0)
        {
            printf("%s: write error !!\n", progname);
            exit(1);
        }
    }

    free(in);
    free(out);

    return 0;
}
